import os
import sys
import threading
import re
import json
import zipfile
import requests
import urllib3.exceptions
from retrying import retry
from tqdm import tqdm
import arrow
from .preprocess import Project


def retry_on_http_error(exception):
    valid_exceptions = [
        requests.exceptions.HTTPError,
        requests.exceptions.ConnectionError,
        urllib3.exceptions.HTTPError,
    ]
    return any([isinstance(exception, e) for e in valid_exceptions])


class Client:
    """Client for communicating with MD.ai backend API.
    Communication is via user access tokens (in MD.ai Hub, Settings -> User Access Tokens).
    """

    def __init__(self, domain="public.md.ai", access_token=None):
        domain_pattern = r"^\w+\.md\.ai(:\d+)?$"
        if not re.match(domain_pattern, domain):
            raise ValueError("domain {} is invalid: should be format *.md.ai".format(domain))

        self.domain = domain
        self.access_token = access_token
        self.session = requests.Session()
        self._test_endpoint()

    def project(self, project_id, path=".", force_download=False, annotations_only=False):
        """Initializes Project class given project id.

        Arguments:
            project_id: hash ID of project.
            path: directory used for data.
        """
        if path == ".":
            print("Using working directory for data.")
        else:
            os.makedirs(path, exist_ok=True)
            print("Using path '{}' for data.".format(path))

        data_manager_kwargs = {
            "domain": self.domain,
            "project_id": project_id,
            "path": path,
            "session": self.session,
            "headers": self._create_headers(),
        }

        annotations_data_manager = ProjectDataManager(
            "annotations", force_download, **data_manager_kwargs
        )

        if not annotations_only:
            images_data_manager = ProjectDataManager(
                "images", force_download, **data_manager_kwargs
            )

        annotations_data_manager.create_data_export_job()

        if not annotations_only:
            images_data_manager.create_data_export_job()

        annotations_data_manager.wait_until_ready()

        if not annotations_only:
            images_data_manager.wait_until_ready()

            p = Project(
                annotations_fp=annotations_data_manager.data_path,
                images_dir=images_data_manager.data_path,
            )
            return p
        else:
            print("No project created. Downloaded annotations only.")
            return None

    def load_model_annotations(self, project_id, dataset_id, model_id, annotations):
        """Method for loading machine learning model results into project as annotations.

        Arguments:
            project_id: hash ID of project.
            dataset_id: hash ID of machine learning model.
            model_id: hash ID of machine learning model.
            annotations: list of annotations to load.
        """
        manager = AnnotationsImportManager(
            domain=self.domain,
            project_id=project_id,
            dataset_id=dataset_id,
            model_id=model_id,
            annotations=annotations,
            session=self.session,
            headers=self._create_headers(),
        )
        manager.create_job()
        manager.wait_until_ready()

    def _create_headers(self):
        headers = {}
        if self.access_token:
            headers["x-access-token"] = self.access_token
        return headers

    def _test_endpoint(self):
        """Checks endpoint for validity and authorization.
        """
        test_endpoint = "https://{}/api/test".format(self.domain)
        r = self.session.get(test_endpoint, headers=self._create_headers())
        if r.status_code == 200:
            print("Successfully authenticated to {}.".format(self.domain))
        else:
            raise Exception("Authorization error. Make sure your access token is valid.")

    @retry(
        retry_on_exception=retry_on_http_error,
        wait_exponential_multiplier=100,
        wait_exponential_max=1000,
        stop_max_attempt_number=10,
    )
    def _gql(self, query, variables=None):
        """Executes GraphQL query.
        """
        gql_endpoint = "https://{}/api/graphql".format(self.domain)
        headers = self._create_headers()
        headers["Accept"] = "application/json"
        headers["Content-Type"] = "application/json"

        data = {"query": query, "variables": variables}
        r = self.session.post(gql_endpoint, headers=headers, json=data)
        if r.status_code != 200:
            r.raise_for_status()

        body = r.json()
        data = body["data"] if "data" in body else None
        errors = body["errors"] if "errors" in body else None

        return data, errors


class ProjectDataManager:
    """Manager for project data exports and downloads.
    """

    def __init__(
        self,
        type,
        force_download,
        domain=None,
        project_id=None,
        path=".",
        session=None,
        headers=None,
    ):
        if type not in ["images", "annotations"]:
            raise ValueError("type must be 'images' or 'annotations'.")
        if not domain:
            raise ValueError("domain is not specified.")
        if not project_id:
            raise ValueError("project_id is not specified.")
        if not os.path.exists(path):
            raise OSError("Path '{}' does not exist.".format(path))

        self.type = type
        self.force_download = force_download

        self.domain = domain
        self.project_id = project_id
        self.path = path
        if session and isinstance(session, requests.Session):
            self.session = session
        else:
            self.session = requests.Session()
        self.headers = headers

        # path for downloaded data
        self.data_path = None
        # ready threading event
        self._ready = threading.Event()

    def create_data_export_job(self):
        """Create data export job through MD.ai API.
        This is an async operation. Status code of 202 indicates successful creation of job.
        """
        endpoint = "https://{}/api/data-export/{}".format(self.domain, self.type)
        params = self._get_data_export_params()
        r = self.session.post(endpoint, json=params, headers=self.headers)
        if r.status_code == 202:
            msg = "Preparing {} export for project {}...".format(self.type, self.project_id)
            print(msg.ljust(100))
            self._check_data_export_job_progress()
        else:
            if r.status_code == 401:
                msg = (
                    "Project {} at domain {}".format(self.project_id, self.domain)
                    + " does not exist or you do not have sufficient permissions for access."
                )
                print(msg)
            self._on_data_export_job_error()

    def wait_until_ready(self):
        self._ready.wait()

    def _get_data_export_params(self):
        if self.type == "images":
            params = {"projectHashId": self.project_id, "exportFormat": "zip"}
        elif self.type == "annotations":
            # TODO: restrict to assigned labelgroup
            params = {
                "projectHashId": self.project_id,
                "labelGroupNum": None,
                "exportFormat": "json",
            }
        return params

    def _check_data_export_job_progress(self):
        """Poll for data export job progress.
        """
        endpoint = "https://{}/api/data-export/{}/progress".format(self.domain, self.type)
        params = self._get_data_export_params()
        r = self.session.post(endpoint, json=params, headers=self.headers)

        try:
            body = r.json()
            status = body["status"]
        except (TypeError, KeyError):
            self._on_data_export_job_error()
            return

        if status == "running":
            try:
                progress = int(body["progress"])
            except (TypeError, ValueError):
                progress = 0
            try:
                time_remaining = int(body["timeRemaining"])
            except (TypeError, ValueError):
                time_remaining = 0

            # print formatted progress info
            if progress > 0 and progress <= 100 and time_remaining > 0:
                if time_remaining > 45:
                    time_remaining_fmt = (
                        arrow.now().shift(seconds=time_remaining).humanize(only_distance=True)
                    )
                else:
                    # arrow humanizes <= 45 to 'in seconds' or 'just now',
                    # so we will opt to be explicit instead.
                    time_remaining_fmt = "in {} seconds".format(time_remaining)
                end_char = "\r" if progress < 100 else "\n"
                msg = "Exporting {} for project {}...{}% (time remaining: {}).".format(
                    self.type, self.project_id, progress, time_remaining_fmt
                )
                print(msg.ljust(100), end=end_char, flush=True)

            # run progress check at 1s intervals so long as status == 'running' and progress < 100
            if progress < 100:
                t = threading.Timer(1.0, self._check_data_export_job_progress)
                t.start()

            return
        elif status == "done":
            self._on_data_export_job_done()
        elif status == "error":
            self._on_data_export_job_error()

    def _on_data_export_job_done(self):
        endpoint = "https://{}/api/data-export/{}/done".format(self.domain, self.type)
        params = self._get_data_export_params()
        r = self.session.post(endpoint, json=params, headers=self.headers)
        try:
            file_key = r.json()["fileKey"]

            if file_key:

                # check cached file
                filepath = os.path.join(self.path, file_key)

                if self.type == "images":
                    data_path = os.path.splitext(filepath)[0]
                elif self.type == "annotations":
                    data_path = filepath

                # if file has been downloaded previously
                if not self.force_download:
                    if os.path.exists(filepath) or os.path.exists(data_path):
                        if self.type == "images":
                            if not os.path.exists(data_path):
                                print("Extracting archive: {}".format(file_key))
                                with zipfile.ZipFile(filepath, "r") as f:
                                    f.extractall(self.path)
                            self.data_path = data_path
                        elif self.type == "annotations":
                            self.data_path = data_path

                        # fire ready threading.Event
                        self._ready.set()
                        print(
                            "Using cached {} data for project {}.".format(
                                self.type, self.project_id
                            )
                        )
                    else:
                        # download in separate thread
                        t = threading.Thread(target=self._download_file, args=(file_key,))
                        t.start()
                else:
                    # download in separate thread
                    t = threading.Thread(target=self._download_file, args=(file_key,))
                    t.start()

        except (TypeError, KeyError):
            self._on_data_export_job_error(type, project_id)

    def _on_data_export_job_error(self):
        endpoint = "https://{}/api/data-export/{}/error".format(self.domain, self.type)
        params = self._get_data_export_params()
        r = self.session.post(endpoint, json=params, headers=self.headers)
        print("Error exporting {} for project {}.".format(self.type, self.project_id))

    def _download_file(self, file_key):
        """Downloads file via signed URL requested from MD.ai API.
        """
        print("Downloading file: {}".format(file_key))
        filepath = os.path.join(self.path, file_key)

        url = "https://{}/api/project-files/signedurl/get?key={}".format(
            self.domain, requests.utils.quote(file_key)
        )

        # stream response so we can display progress bar
        r = requests.get(url, stream=True, headers=self.headers)

        # total size in bytes
        total_size = int(r.headers.get("content-length", 0))
        block_size = 32 * 1024
        wrote = 0
        with open(filepath, "wb") as f:
            with tqdm(total=total_size, unit="B", unit_scale=True, unit_divisor=1024) as pbar:
                for chunk in r.iter_content(block_size):
                    f.write(chunk)
                    wrote = wrote + len(chunk)
                    pbar.update(block_size)
        if total_size != 0 and wrote != total_size:
            raise IOError("Error downloading file {}.".format(file_key))

        if self.type == "images":
            # unzip archive
            print("Extracting archive: {}".format(file_key))
            with zipfile.ZipFile(filepath, "r") as f:
                f.extractall(self.path)
            # images directory should have same name as archive file
            self.data_path = os.path.splitext(filepath)[0]
        elif self.type == "annotations":
            self.data_path = filepath

        # fire ready threading.Event
        self._ready.set()
        print("Success: {} data for project {} ready.".format(self.type, self.project_id))


class AnnotationsImportManager:
    """Manager for importing annotations.
    """

    def __init__(
        self,
        domain=None,
        project_id=None,
        dataset_id=None,
        model_id=None,
        annotations=None,
        session=None,
        headers=None,
    ):
        if not domain:
            raise ValueError("domain is not specified.")
        if not project_id:
            raise ValueError("project_id is not specified.")

        self.domain = domain
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.model_id = model_id
        self.annotations = annotations
        if session and isinstance(session, requests.Session):
            self.session = session
        else:
            self.session = requests.Session()
        self.headers = headers

        self.job_id = None

        # ready threading event
        self._ready = threading.Event()

    def create_job(self):
        """Create annotations import job through MD.ai API.
        This is an async operation. Status code of 202 indicates successful creation of job.
        """
        endpoint = "https://{}/api/data-import/annotations".format(self.domain)
        params = {
            "projectHashId": self.project_id,
            "datasetHashId": self.dataset_id,
            "modelHashId": self.model_id,
            "annotations": self.annotations,
        }
        r = self.session.post(endpoint, json=params, headers=self.headers)
        if r.status_code == 202:
            self.job_id = r.json()["jobId"]
            msg = (
                "Importing annotations into project {}, ".format(self.project_id)
                + "dataset {}, ".format(self.dataset_id)
                + "model {}...".format(self.model_id)
            )
            print(msg.ljust(100))
            self._check_job_progress()
        else:
            if r.status_code == 401:
                msg = (
                    "Project, dataset, or model does not exist, "
                    + "or you do not have sufficient permissions."
                )
                print(msg)
            self._on_job_error()

    def wait_until_ready(self):
        self._ready.wait()

    def _check_job_progress(self):
        """Poll for annotations import job progress.
        """
        endpoint = "https://{}/api/data-import/annotations/progress".format(self.domain)
        params = {"projectHashId": self.project_id, "jobId": self.job_id}
        r = self.session.post(endpoint, json=params, headers=self.headers)

        try:
            body = r.json()
            status = body["status"]
        except (TypeError, KeyError):
            self._on_job_error()
            return

        if status == "running":
            try:
                progress = int(body["progress"])
            except (TypeError, ValueError):
                progress = 0
            try:
                time_remaining = int(body["timeRemaining"])
            except (TypeError, ValueError):
                time_remaining = 0

            # print formatted progress info
            if progress > 0 and progress <= 100 and time_remaining > 0:
                if time_remaining > 45:
                    time_remaining_fmt = (
                        arrow.now().shift(seconds=time_remaining).humanize(only_distance=True)
                    )
                else:
                    # arrow humanizes <= 45 to 'in seconds' or 'just now',
                    # so we will opt to be explicit instead.
                    time_remaining_fmt = "in {} seconds".format(time_remaining)
                end_char = "\r" if progress < 100 else "\n"
                msg = "Annotations import for project {}...{}% (time remaining: {}).".format(
                    self.project_id, progress, time_remaining_fmt
                )
                print(msg.ljust(100), end=end_char, flush=True)

            # run progress check at 1s intervals so long as status == 'running' and progress < 100
            if progress < 100:
                t = threading.Timer(1.0, self._check_job_progress)
                t.start()

            return
        elif status == "done":
            self._on_job_done()
            # fire ready threading.Event
            self._ready.set()
        elif status == "error":
            self._on_job_error()

    def _on_job_done(self):
        endpoint = "https://{}/api/data-export/annotations/done".format(self.domain)
        params = {"projectHashId": self.project_id, "jobId": self.job_id}
        r = self.session.post(endpoint, json=params, headers=self.headers)
        print("Successfully imported annotations into project {}.".format(self.project_id))

    def _on_data_export_job_error(self):
        endpoint = "https://{}/api/data-export/annotations/error".format(self.domain)
        params = {"projectHashId": self.project_id, "jobId": self.job_id}
        r = self.session.post(endpoint, json=params, headers=self.headers)
        print("Error importing annotations into project {}.".format(self.project_id))




import pytest
from mdai import visualize
import numpy as np

# TODO: test load_dicom_image (with RGB option or not)


def test_visualize(p):
    labels_dict = {
        "L_egJRyg": 1,  # bounding box
        "L_MgevP2": 2,  # polygon
        "L_D21YL2": 3,  # freeform
        "L_lg7klg": 4,  # line
        "L_eg69RZ": 5,  # location
        "L_GQoaJg": 6,  # global_image
        "L_JQVWjZ": 7,  # global_series
        "L_3QEOpg": 8,  # global_exam
    }

    p.set_labels_dict(labels_dict)

    ct_dataset = p.get_dataset_by_id("D_qGQdpN")
    ct_dataset.prepare()

    # image with multiple annotations
    image_id = ct_dataset.get_image_ids()[7]

    grey_image = visualize.load_dicom_image(image_id)
    rgb_image = visualize.load_dicom_image(image_id, to_RGB=True)
    scaled_image_1 = visualize.load_dicom_image(image_id, rescale=True)

    assert np.amax(grey_image) == 701
    assert np.amax(scaled_image_1) == 255

    assert grey_image.shape == (256, 256)
    assert rgb_image.shape == (256, 256, 3)

    scaled_image_2, gt_class_id, gt_bbox, gt_mask = visualize.get_image_ground_truth(
        image_id, ct_dataset
    )

    assert len(gt_class_id) == len(gt_class_id)
    assert gt_mask.shape == (256, 256, 5)



import pytest
from mdai.utils.common_utils import train_test_split


def test_project(p):

    # label groups
    label_groups = p.get_label_groups()

    assert label_groups[0].id == "G_L3dP31"
    assert label_groups[1].id == "G_WVRrVJ"


def test_dataset(p):

    # two datasets
    datasets = p.get_datasets()
    assert len(datasets) == 2

    labels_dict = {
        "L_egJRyg": 1,  # bounding box
        "L_MgevP2": 2,  # polygon
        "L_D21YL2": 3,  # freeform
        "L_lg7klg": 4,  # line
        "L_eg69RZ": 5,  # location
        "L_GQoaJg": 6,  # global_image
        "L_JQVWjZ": 7,  # global_series
        "L_3QEOpg": 8,  # global_exam
    }
    p.set_labels_dict(labels_dict)

    assert p.get_label_id_annotation_mode("L_MgevP2") == "polygon"
    assert p.get_label_id_annotation_mode("L_3QEOpg") is None

    ct_dataset = p.get_dataset_by_id("D_qGQdpN")
    ct_dataset.prepare()

    xray_dataset = p.get_dataset_by_id("D_0Z4nDG")
    xray_dataset.prepare()

    assert ct_dataset.classes_dict == xray_dataset.classes_dict

    image_ids = ct_dataset.get_image_ids()
    assert len(image_ids) == len(ct_dataset.imgs_anns_dict.keys())

    image_id = ct_dataset.get_image_ids()[7]

    ann_mode = [
        (ct_dataset.label_id_to_class_annotation_mode(ann["labelId"]), ann["labelId"])
        for ann in ct_dataset.imgs_anns_dict[image_id]
    ]

    assert ann_mode == [
        ("line", "L_lg7klg"),
        ("polygon", "L_MgevP2"),
        ("freeform", "L_D21YL2"),
        ("bbox", "L_egJRyg"),
        ("location", "L_eg69RZ"),
        (None, "L_3QEOpg"),
    ]

    train_ds, valid_ds = train_test_split(ct_dataset, shuffle=False, validation_split=0.2)

    assert len(train_ds.get_image_ids()) == 9
    assert len(valid_ds.get_image_ids()) == 3



"""MD.ai Python client library."""

__version__ = "0.1.0"

from . import preprocess
from .client import Client
from .utils import common_utils

try:
    CAN_VISUALIZE = True
    from . import visualize
except ImportError:
    # matplotlib backend missing or cannot be loaded
    CAN_VISUALIZE = False



import random
import copy


def hex2rgb(h):
    """Convert Hex color encoding to RGB color"""
    h = h.lstrip("#")
    return tuple(int(h[i : i + 2], 16) for i in (0, 2, 4))


def train_test_split(dataset, shuffle=True, validation_split=0.1):
    """
    Split image ids into training and validation sets.
    TODO: Need to update dataset.dataset_data!
    TODO: What to set for images_dir for combined dataset?
    """
    if validation_split < 0.0 or validation_split > 1.0:
        raise ValueError("{} is not a valid split ratio.".format(validation_split))

    image_ids_list = dataset.get_image_ids()
    if shuffle:
        sorted(image_ids_list)
        random.seed(42)
        random.shuffle(image_ids_list)

    split_index = int((1 - validation_split) * len(image_ids_list))
    train_image_ids = image_ids_list[:split_index]
    valid_image_ids = image_ids_list[split_index:]

    def filter_by_ids(ids, imgs_anns_dict):
        return {x: imgs_anns_dict[x] for x in ids}

    train_dataset = copy.deepcopy(dataset)
    train_dataset.id = dataset.id + "-TRAIN"

    valid_dataset = copy.deepcopy(dataset)
    valid_dataset.id = dataset.id + "-VALID"

    imgs_anns_dict = dataset.imgs_anns_dict

    train_imgs_anns_dict = filter_by_ids(train_image_ids, imgs_anns_dict)
    valid_imgs_anns_dict = filter_by_ids(valid_image_ids, imgs_anns_dict)

    train_dataset.image_ids = train_image_ids
    valid_dataset.image_ids = valid_image_ids

    train_dataset.imgs_anns_dict = train_imgs_anns_dict
    valid_dataset.imgs_anns_dict = valid_imgs_anns_dict

    print(
        "Num of instances for training set: %d, validation set: %d"
        % (len(train_image_ids), len(valid_image_ids))
    )
    return train_dataset, valid_dataset
